@@ -63,7 +63,7 @@
 #define MINVERSION 6 /* API compatible, ABI may change, functional
 		      * enhancements only, consumer can be left unchanged if
 		      * enhancements are not considered */
-#define PATCHLEVEL 4 /* API / ABI compatible, no functional changes, no
+#define PATCHLEVEL 5 /* API / ABI compatible, no functional changes, no
 		      * enhancements, bug fixes only */
 
 /* remove once in if_alg.h */
@@ -95,6 +95,7 @@ static ssize_t _kcapi_common_send_meta(struct kcapi_handle *handle,
 	ssize_t ret = -EINVAL;
 	char *buffer = NULL;
 	volatile void *_buffer = NULL;
+	int errsv = 0;
 
 	/* plaintext / ciphertext data */
 	struct cmsghdr *header = NULL;
@@ -159,6 +160,7 @@ static ssize_t _kcapi_common_send_meta(struct kcapi_handle *handle,
 	}
 
 	ret = sendmsg(handle->opfd, &msg, flags);
+	errsv = errno;
 
 	memset(buffer, 0, bufferlen);
 	/* magic to convince GCC to memset the buffer */
@@ -166,7 +168,7 @@ static ssize_t _kcapi_common_send_meta(struct kcapi_handle *handle,
 	if (_buffer)
 		_buffer = '\0';
 	free(buffer);
-	return ret;
+	return (ret >= 0) ? ret : -errsv;
 }
 
 static inline ssize_t _kcapi_common_send_data(struct kcapi_handle *handle,
@@ -174,6 +176,7 @@ static inline ssize_t _kcapi_common_send_data(struct kcapi_handle *handle,
 				       unsigned int flags)
 {
 	struct msghdr msg;
+	ssize_t ret = 0;
 
 	msg.msg_name = NULL;
 	msg.msg_namelen = 0;
@@ -183,7 +186,8 @@ static inline ssize_t _kcapi_common_send_data(struct kcapi_handle *handle,
 	msg.msg_iov = iov;
 	msg.msg_iovlen = iovlen;
 
-	return sendmsg(handle->opfd, &msg, flags);
+	ret = sendmsg(handle->opfd, &msg, flags);
+	return (ret >= 0) ? ret : -errno;
 }
 
 static inline ssize_t _kcapi_common_vmsplice_data(struct kcapi_handle *handle,
@@ -206,7 +210,8 @@ static inline ssize_t _kcapi_common_vmsplice_data(struct kcapi_handle *handle,
 	if ((size_t)ret != inlen)
 		fprintf(stderr, "vmsplice: not all data received by kernel (data recieved: %ld -- data sent: %lu)\n",
 			(long)ret, (unsigned long)inlen);
-	return splice(handle->pipes[0], NULL, handle->opfd, NULL, ret, flags);
+	ret = splice(handle->pipes[0], NULL, handle->opfd, NULL, ret, flags);
+	return (ret >= 0) ? ret : -errno;
 }
 
 static inline ssize_t _kcapi_common_recv_data(struct kcapi_handle *handle,
@@ -214,6 +219,7 @@ static inline ssize_t _kcapi_common_recv_data(struct kcapi_handle *handle,
 {
 	struct msghdr msg;
 	ssize_t ret = 0;
+	int errsv = 0;
 
 	msg.msg_name = NULL;
 	msg.msg_namelen = 0;
@@ -224,25 +230,30 @@ static inline ssize_t _kcapi_common_recv_data(struct kcapi_handle *handle,
 	msg.msg_iovlen = iovlen;
 
 	ret = recvmsg(handle->opfd, &msg, 0);
-	if (msg.msg_flags & MSG_TRUNC)
-		return -ENOMEM;
-	return ret;
+	errsv = errno;
+	if (msg.msg_flags & MSG_TRUNC) {
+		fprintf(stderr, "recvmsg: processed data was truncated by kernel (only %lu bytes processed)\n", (unsigned long)ret);
+		return -EMSGSIZE;
+	}
+	return (ret >= 0) ? ret : -errsv;
 }
 
 static inline ssize_t _kcapi_common_read_data(struct kcapi_handle *handle,
 					      unsigned char *out, size_t outlen)
 {
-	return read(handle->opfd, out, outlen);
+	ssize_t ret = 0;
+
+	ret = read(handle->opfd, out, outlen);
+	return (ret >= 0) ? ret : -errno;
 }
 
 static inline int _kcapi_common_setkey(struct kcapi_handle *handle,
 				       const unsigned char *key, size_t keylen)
 {
-	if (setsockopt(handle->tfmfd, SOL_ALG, ALG_SET_KEY,
-		       key, keylen) == -1)
-		return -EINVAL;
+	int ret = 0;
 
-	return 0;
+	ret = setsockopt(handle->tfmfd, SOL_ALG, ALG_SET_KEY, key, keylen);
+	return (ret >= 0) ? ret : -errno;
 }
 
 static inline int _kcapi_common_setiv(struct kcapi_handle *handle,
@@ -266,6 +277,7 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 				  int drivername)
 {
 	int ret = -EFAULT;
+	int errsv = 0;
 
 	/* NETLINK_CRYPTO specific */
 	char buf[4096];
@@ -305,26 +317,30 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 	sd =  socket(AF_NETLINK, SOCK_RAW, NETLINK_CRYPTO);
 	if (sd < 0) {
 		perror("Netlink error: cannot open netlink socket");
-		return -EFAULT;
+		return -errno;
 	}
 	memset(&nl, 0, sizeof(nl));
 	nl.nl_family = AF_NETLINK;
 	if (bind(sd, (struct sockaddr*)&nl, sizeof(nl)) < 0) {
+		errsv = errno;
 		perror("Netlink error: cannot bind netlink socket");
 		goto out;
 	}
 	/* sanity check that netlink socket was successfully opened */
 	addr_len = sizeof(nl);
 	if (getsockname(sd, (struct sockaddr*)&nl, &addr_len) < 0) {
+		errsv = errno;
 		perror("Netlink error: cannot getsockname");
 		goto out;
 	}
 	if (addr_len != sizeof(nl)) {
+		errsv = errno;
 		fprintf(stderr, "Netlink error: wrong address length %d\n",
 			addr_len);
 		goto out;
 	}
 	if (nl.nl_family != AF_NETLINK) {
+		errsv = errno;
 		fprintf(stderr, "Netlink error: wrong address family %d\n",
 			nl.nl_family);
 		goto out;
@@ -340,6 +356,7 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 	memset(&nl, 0, sizeof(nl));
 	nl.nl_family = AF_NETLINK;
 	if (sendmsg(sd, &msg, 0) < 0) {
+		errsv = errno;
 		perror("Netlink error: sendmsg failed");
 		goto out;
 	}
@@ -351,18 +368,18 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 		if (ret < 0) {
 			if (errno == EINTR || errno == EAGAIN)
 				continue;
+			errsv = errno;
 			perror("Netlink error: netlink receive error");
-			ret = -EFAULT;
 			goto out;
 		}
 		if (ret == 0) {
+			errsv = errno;
 			fprintf(stderr, "Netlink error: no data\n");
-			ret = -EFAULT;
 			goto out;
 		}
 		if ((size_t)ret > sizeof(buf)) {
+			errsv = errno;
 			perror("Netlink error: received too much data\n");
-			ret = -EFAULT;
 			goto out;
 		}
 		break;
@@ -438,7 +455,7 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 
 out:
 	close(sd);
-	return ret;
+	return (errsv) ? -errsv : ret;
 }
 
 static int _kcapi_common_getinfo(struct kcapi_handle *handle,
@@ -455,6 +472,7 @@ static int _kcapi_handle_init(struct kcapi_handle *handle,
 {
 	struct sockaddr_alg sa;
 	int ret;
+	int errsv = 0;
 
 	memset(handle, 0, sizeof(struct kcapi_handle));
 
@@ -468,29 +486,33 @@ static int _kcapi_handle_init(struct kcapi_handle *handle,
 		return -EOPNOTSUPP;
 
 	if (bind(handle->tfmfd, (struct sockaddr *)&sa, sizeof(sa)) == -1) {
+		errsv = errno;
 		perror("AF_ALG: bind failed");
 		close(handle->tfmfd);
 		handle->tfmfd = -1;
-		return -ENOENT;
+		return -errsv;
 	}
 
 	handle->opfd = accept(handle->tfmfd, NULL, 0);
 	if (handle->opfd == -1) {
+		errsv = errno;
 		perror("AF_ALG: accept failed");
 		close(handle->tfmfd);
 		handle->tfmfd = -1;
-		return -EINVAL;
+		return -errsv;
 	}
 
 	ret = pipe(handle->pipes);
 	if (ret) {
+		errsv = errno;
 		close(handle->tfmfd);
 		close(handle->opfd);
-		return ret;
+		return -errsv;
 	}
 
 	ret = _kcapi_common_getinfo(handle, ciphername);
 	if(ret) {
+		errsv = errno;
 		fprintf(stderr, "NETLINK_CRYPTO: cannot obtain cipher information for %s (is required crypto_user.c patch missing? see documentation)\n",
 		       ciphername);
 		close(handle->tfmfd);
@@ -498,7 +520,7 @@ static int _kcapi_handle_init(struct kcapi_handle *handle,
 		close(handle->pipes[0]);
 		close(handle->pipes[1]);
 	}
-	return ret;
+	return (errsv) ? -errsv : ret;
 }
 
 static inline int _kcapi_handle_destroy(struct kcapi_handle *handle)
@@ -573,20 +595,22 @@ unsigned int kcapi_version(void)
 }
 
 /**
- * kcapi_pad_iv() - realign the key as necessary for cipher
+ * kcapi_pad_iv() - realign the IV as necessary for cipher
  * @handle: cipher handle
  * @iv: current IV buffer - input
  * @ivlen: length of IV buffer - input
  * @newiv: buffer of aligned IV - output
  * @newivlen: length of newly aligned IV - output
  *
+ * The function pads the least significant bits of the provided IV up to the
+ * block size of the cipher with zeros. In case the provided IV is longer than
+ * the block size, the least significant bits are truncated to the block size.
+ *
  * The function allocates memory for @newiv in case the return code indicates
  * success. The consumer must free the memory after use.
  *
  * Return: 0 for success;
- *	   -ERANGE when the provided IV already satisfies the
- *	   the alignment requirement;
- *	   < 0 for any other errors
+ *	   < 0 for any errors
  */
 int kcapi_pad_iv(struct kcapi_handle *handle,
 		 const unsigned char *iv, size_t ivlen,
@@ -594,14 +618,14 @@ int kcapi_pad_iv(struct kcapi_handle *handle,
 {
 	unsigned char *niv = NULL;
 	unsigned int nivlen = handle->info.ivsize;
+	unsigned int copylen = (ivlen > nivlen) ? nivlen : ivlen;
+	int ret = 0;
 
-	if (nivlen == ivlen)
-		return -ERANGE;
-
-	niv = calloc(1, nivlen);
-	if (!niv)
-		return -ENOMEM;
-	memcpy(niv, iv, nivlen);
+	ret = posix_memalign((void *)&niv, 16, nivlen);
+	if (ret)
+		return -ret;
+	memset(niv, 0, nivlen);
+	memcpy(niv, iv, copylen);
 
 	*newiv = niv;
 	*newivlen = nivlen;
@@ -650,6 +674,12 @@ int kcapi_cipher_destroy(struct kcapi_handle *handle)
  * @key: key buffer - input
  * @keylen: length of key buffer - input
  *
+ * With this function, the caller sets the key for subsequent encryption or
+ * decryption operations.
+ *
+ * After the caller provided the key, the caller may securely destroy the key
+ * as it is now maintained by the kernel.
+ *
  * Return: 0 upon success;
  *	   < 0 in case of error
  */
@@ -668,6 +698,11 @@ int kcapi_cipher_setkey(struct kcapi_handle *handle,
  * This function requires IV to be exactly IV size. The function verifies
  * the IV size to avoid unnecessary kernel round trips.
  *
+ * Note - the caller must keep the iv buffer for every kcapi_cipher_encrypt() or
+ * kcapi_cipher_decrypt() operation, or until the last invocation of
+ * kcapi_cipher_stream_init_enc() or kcapi_cipher_stream_init_dec() is
+ * performed.
+ *
  * Return: 0 upon success;
  *	   < 0 in case of an error
  *
@@ -690,6 +725,10 @@ int kcapi_cipher_setiv(struct kcapi_handle *handle,
  * ciphertext pointers. That would mean that after the encryption operation,
  * the plaintext is overwritten with the ciphertext.
  *
+ * The memory should be aligned at the page boundary using
+ * posix_memalign(PAGE_SIZE), If it is not aligned at the page boundary,
+ * the vmsplice call may not send all data to the kernel.
+ *
  * Return: number of bytes encrypted upon success;
  *	   < 0 in case of error with errno set
  */
@@ -717,21 +756,19 @@ ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
 	}
 
 	iov.iov_base = (void*)(uintptr_t)in;
-#if 0
+
 	/*
 	 * Using two syscalls with memcpy is faster than four syscalls
 	 * without memcpy below the given threshold.
 	 */
-	/* TODO make heuristic when one syscall is slower than four syscalls */
-	if (inlen < (1<<15)) {
+	if (inlen <= (1<<13)) {
 		iov.iov_len = inlen;
 		ret = _kcapi_common_send_meta(handle, &iov, 1, ALG_OP_ENCRYPT,
 					      0);
 		iov.iov_base = (void*)(uintptr_t)out;
 		iov.iov_len = outlen;
 		return _kcapi_common_recv_data(handle, &iov, 1);
-}
-#endif
+	}
 
 	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_ENCRYPT, 0);
 	if (0 > ret)
@@ -765,6 +802,10 @@ ssize_t kcapi_cipher_encrypt(struct kcapi_handle *handle,
  * ciphertext pointers. That would mean that after the encryption operation,
  * the ciphertext is overwritten with the plaintext.
  *
+ * The memory should be aligned at the page boundary using
+ * posix_memalign(PAGE_SIZE), If it is not aligned at the page boundary,
+ * the vmsplice call may not send all data to the kernel.
+ *
  * Return: number of bytes decrypted upon success;
  *	   < 0 in case of error with errno set
  */
@@ -798,13 +839,12 @@ ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 	}
 
 	iov.iov_base = (void*)(uintptr_t)in;
-#if 0
+
 	/*
 	 * Using two syscalls with memcpy is faster than four syscalls
 	 * without memcpy below the given threshold.
 	 */
-	/* TODO make heuristic when one syscall is slower than four syscalls */
-	if (inlen < (1<<15)) {
+	if (inlen <= (1<<13)) {
 		iov.iov_len = inlen;
 		ret = _kcapi_common_send_meta(handle, &iov, 1, ALG_OP_DECRYPT,
 					      0);
@@ -812,7 +852,6 @@ ssize_t kcapi_cipher_decrypt(struct kcapi_handle *handle,
 		iov.iov_len = outlen;
 		return _kcapi_common_recv_data(handle, &iov, 1);
 	}
-#endif
 
 	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_DECRYPT, 0);
 	if (0 > ret)
@@ -1025,6 +1064,12 @@ int kcapi_aead_destroy(struct kcapi_handle *handle)
  * @key: key buffer - input
  * @keylen: length of key buffer - input
  *
+ * With this function, the caller sets the key for subsequent encryption or
+ * decryption operations.
+ *
+ * After the caller provided the key, the caller may securely destroy the key
+ * as it is now maintained by the kernel.
+ *
  * Return: 0 upon success;
  *	   < 0 in case of error
  */
@@ -1044,6 +1089,11 @@ int kcapi_aead_setkey(struct kcapi_handle *handle,
  *
  * This function requires IV to be exactly IV size. The function verifies
  * the IV size to avoid unnecessary kernel round trips.
+ *
+ * Note - the caller must keep the iv buffer for every kcapi_aead_encrypt() or
+ * kcapi_aead_decrypt() operation, or until the last invocation of
+ * kcapi_aead_stream_init_enc() or kcapi_aead_stream_init_dec() is performed.
+ * This function only retains a pointer to the user-provided IV buffer.
  */
 int kcapi_aead_setiv(struct kcapi_handle *handle,
 		     const unsigned char *iv, size_t ivlen)
@@ -1077,6 +1127,12 @@ int kcapi_aead_settaglen(struct kcapi_handle *handle, size_t taglen)
  * kcapi_aead_setassoclen() - Set authentication data size
  * @handle: cipher handle - input
  * @assoclen: length of associated data length
+ *
+ * The associated data is retained in the cipher handle. During initialization
+ * of a cipher handle, it is sent to the kernel. The kernel cipher
+ * implementations may verify the appropriateness of the authentication
+ * data size and may return an error during initialization if the
+ * authentication size is not considered appropriate.
  */
 void kcapi_aead_setassoclen(struct kcapi_handle *handle, size_t assoclen)
 {
@@ -1100,6 +1156,10 @@ void kcapi_aead_setassoclen(struct kcapi_handle *handle, size_t assoclen)
  * ciphertext pointers. That would mean that after the encryption operation,
  * the plaintext is overwritten with the ciphertext.
  *
+ * The memory should be aligned at the page boundary using
+ * posix_memalign(PAGE_SIZE), If it is not aligned at the page boundary,
+ * the vmsplice call may not send all data to the kernel.
+ *
  * After invoking this function the caller should use
  * kcapi_aead_getdata() to obtain the resulting ciphertext and authentication
  * tag references.
@@ -1114,7 +1174,6 @@ ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 {
 	struct iovec iov[2];
 	ssize_t ret = 0;
-	size_t len = 0;
 
 	/* require properly sized output data size */
 	if (outlen < _kcapi_aead_encrypt_outlen(handle, inlen,
@@ -1128,33 +1187,38 @@ ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 	}
 
 #if 0
-	/* using two syscall */
-	/* TODO make heuristic when one syscall is slower than four syscalls */
-	if (assoc && handle->aead.assoclen) {
-		iov[0].iov_base = (void*)(uintptr_t)assoc;
-		iov[0].iov_len = handle->aead.assoclen;
-		iov[1].iov_base = (void*)(uintptr_t)in;
-		iov[1].iov_len = inlen;
-		ret = _kcapi_common_send_meta(handle, &iov[0], 2,
-					      ALG_OP_ENCRYPT, 0);
-	} else {
-		iov[0].iov_base = (void*)(uintptr_t)in;
-		iov[0].iov_len = inlen;
-		ret = _kcapi_common_send_meta(handle, &iov[0], 1,
-					      ALG_OP_ENCRYPT, 0);
-	}
-	if (0 > ret)
-		return ret;
-
-	iov[0].iov_base = (void*)(uintptr_t)out;
-	iov[0].iov_len = outlen;
-	ret = _kcapi_common_recv_data(handle, &iov[0], 1);
+	if (inlen + handle->aead.assoclen >= (1<<17)) {
 #endif
+		if (assoc && handle->aead.assoclen) {
+			iov[0].iov_base = (void*)(uintptr_t)assoc;
+			iov[0].iov_len = handle->aead.assoclen;
+			iov[1].iov_base = (void*)(uintptr_t)in;
+			iov[1].iov_len = inlen;
+			ret = _kcapi_common_send_meta(handle, &iov[0], 2,
+						ALG_OP_ENCRYPT, 0);
+		} else {
+			iov[0].iov_base = (void*)(uintptr_t)in;
+			iov[0].iov_len = inlen;
+			ret = _kcapi_common_send_meta(handle, &iov[0], 1,
+						ALG_OP_ENCRYPT, 0);
+		}
+		if (0 > ret)
+			return ret;
+		iov[0].iov_base = (void*)(uintptr_t)out;
+		iov[0].iov_len = outlen;
+		return _kcapi_common_recv_data(handle, &iov[0], 1);
+
+#if 0
+	}
+
+	/* using separate pages slows down splice below the speed of sendmsg */
 	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_ENCRYPT,
 				      MSG_MORE);
 	if (0 > ret)
 		return ret;
 	if (assoc && handle->aead.assoclen) {
+		size_t len = 0;
+
 		iov[0].iov_base = (void*)(uintptr_t)assoc;
 		iov[0].iov_len = handle->aead.assoclen;
 		len = handle->aead.assoclen;
@@ -1163,6 +1227,8 @@ ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 		len += inlen;
 		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 2, len, 0);
 	} else {
+		size_t len = 0;
+
 		iov[0].iov_base = (void*)(uintptr_t)in;
 		iov[0].iov_len = inlen;
 		len = inlen;
@@ -1171,11 +1237,16 @@ ssize_t kcapi_aead_encrypt(struct kcapi_handle *handle,
 	if (0 > ret)
 		return ret;
 
-	ret = _kcapi_common_read_data(handle, out, outlen);
+	iov[0].iov_base = (void*)(uintptr_t)out;
+	iov[0].iov_len = outlen;
+	ret = _kcapi_common_recv_data(handle, &iov[0], 1);
+	if (ret < 0)
+		return ret;
 	if ((ret < (ssize_t)handle->aead.taglen))
 		return -E2BIG;
 
 	return ret;
+#endif
 }
 
 /**
@@ -1236,6 +1307,10 @@ void kcapi_aead_getdata(struct kcapi_handle *handle,
  * ciphertext pointers. That would mean that after the encryption operation,
  * the ciphertext is overwritten with the plaintext.
  *
+ * The memory should be aligned at the page boundary using
+ * posix_memalign(PAGE_SIZE), If it is not aligned at the page boundary,
+ * the vmsplice call may not send all data to the kernel.
+ *
  * To catch authentication errors (i.e. integrity violations) during the
  * decryption operation, the errno of this call shall be checked for EBADMSG.
  * If this function returns < 0 and errno is set to EBADMSG, an authentication
@@ -1251,7 +1326,6 @@ ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 {
 	struct iovec iov[3];
 	ssize_t ret = 0;
-	size_t len = 0;
 	unsigned int bs = handle->info.blocksize;
 
 	/* require properly sized output data size */
@@ -1264,60 +1338,66 @@ ssize_t kcapi_aead_decrypt(struct kcapi_handle *handle,
 	}
 
 #if 0
-	/* using two syscall */
-	/* TODO make heuristic when one syscall is slower than four syscalls */
-	if (assoc && handle->aead.assoclen) {
-		iov[0].iov_base = (void*)(uintptr_t)assoc;
-		iov[0].iov_len = handle->aead.assoclen;
-		iov[1].iov_base = (void*)(uintptr_t)in;
-		iov[1].iov_len = inlen;
-		iov[2].iov_base = (void*)(uintptr_t)tag;
-		iov[2].iov_len = handle->aead.taglen;
-		ret = _kcapi_common_send_meta(handle, &iov[0], 3,
-					      ALG_OP_DECRYPT, 0);
+	if (inlen + handle->aead.assoclen + handle->aead.taglen >= (1<<13s)) {
+#endif
+		if (assoc && handle->aead.assoclen) {
+			iov[0].iov_base = (void*)(uintptr_t)assoc;
+			iov[0].iov_len = handle->aead.assoclen;
+			iov[1].iov_base = (void*)(uintptr_t)in;
+			iov[1].iov_len = inlen;
+			iov[2].iov_base = (void*)(uintptr_t)tag;
+			iov[2].iov_len = handle->aead.taglen;
+			ret = _kcapi_common_send_meta(handle, &iov[0], 3,
+						ALG_OP_DECRYPT, 0);
+		} else {
+			iov[0].iov_base = (void*)(uintptr_t)in;
+			iov[0].iov_len = inlen;
+			iov[1].iov_base = (void*)(uintptr_t)tag;
+			iov[1].iov_len = handle->aead.taglen;
+			ret = _kcapi_common_send_meta(handle, &iov[0], 2,
+						ALG_OP_DECRYPT, 0);
+		}
+		if (0 > ret)
+			return ret;
+#if 0
+	/* using separate pages slows down splice below the speed of sendmsg */
 	} else {
-		iov[0].iov_base = (void*)(uintptr_t)in;
-		iov[0].iov_len = inlen;
-		iov[1].iov_base = (void*)(uintptr_t)tag;
-		iov[1].iov_len = handle->aead.taglen;
-		ret = _kcapi_common_send_meta(handle, &iov[0], 2,
-					      ALG_OP_DECRYPT, 0);
+		size_t len = 0;
+
+		ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_DECRYPT,
+					MSG_MORE);
+		if (0 > ret)
+			return ret;
+		if (assoc && handle->aead.assoclen) {
+			iov[0].iov_base = (void*)(uintptr_t)assoc;
+			iov[0].iov_len = handle->aead.assoclen;
+			len = handle->aead.assoclen;
+			iov[1].iov_base = (void*)(uintptr_t)in;
+			iov[1].iov_len = inlen;
+			len += inlen;
+			iov[2].iov_base = (void*)(uintptr_t)tag;
+			iov[2].iov_len = handle->aead.taglen;
+			len += handle->aead.taglen;
+			ret = _kcapi_common_vmsplice_data(handle, &iov[0], 3,
+							  len, 0);
+		} else {
+			iov[0].iov_base = (void*)(uintptr_t)in;
+			iov[0].iov_len = inlen;
+			len = inlen;
+			iov[1].iov_base = (void*)(uintptr_t)tag;
+			iov[1].iov_len = handle->aead.taglen;
+			len += handle->aead.taglen;
+			ret = _kcapi_common_vmsplice_data(handle, &iov[0], 2,
+							  len, 0);
+		}
+		if (0 > ret)
+			return ret;
 	}
-	if (0 > ret)
-		return ret;
+#endif
 
 	iov[0].iov_base = (void*)(uintptr_t)out;
 	iov[0].iov_len = outlen;
-	ret = _kcapi_common_recv_data(handle, &iov[0], 1);
-#endif
-	ret = _kcapi_common_send_meta(handle, NULL, 0, ALG_OP_DECRYPT,
-				      MSG_MORE);
-	if (0 > ret)
-		return ret;
-	if (assoc && handle->aead.assoclen) {
-		iov[0].iov_base = (void*)(uintptr_t)assoc;
-		iov[0].iov_len = handle->aead.assoclen;
-		len = handle->aead.assoclen;
-		iov[1].iov_base = (void*)(uintptr_t)in;
-		iov[1].iov_len = inlen;
-		len += inlen;
-		iov[2].iov_base = (void*)(uintptr_t)tag;
-		iov[2].iov_len = handle->aead.taglen;
-		len += handle->aead.taglen;
-		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 3, len, 0);
-	} else {
-		iov[0].iov_base = (void*)(uintptr_t)in;
-		iov[0].iov_len = inlen;
-		len = inlen;
-		iov[1].iov_base = (void*)(uintptr_t)tag;
-		iov[1].iov_len = handle->aead.taglen;
-		len += handle->aead.taglen;
-		ret = _kcapi_common_vmsplice_data(handle, &iov[0], 2, len, 0);
-	}
-	if (0 > ret)
-		return ret;
-
-	return _kcapi_common_read_data(handle, out, outlen);
+	return _kcapi_common_recv_data(handle, &iov[0], 1);
 }
 
 /**
@@ -1571,12 +1651,14 @@ int kcapi_aead_ccm_nonce_to_iv(const unsigned char *nonce, size_t noncelen,
 {
 	unsigned char *newiv = NULL;
 	unsigned char l = 16 - 2 - noncelen;
+	int ret = 0;
 
 	if (noncelen > 16 - 2)
 		return -EINVAL;
-	newiv = calloc(1, 16);
-	if (!newiv)
-		return -ENOMEM;
+	ret = posix_memalign((void *)&newiv, 16, 16);
+	if (ret)
+		return -ret;
+	memset(newiv, 0, 16);
 	newiv[0] = l;
 	memcpy(newiv + 1, nonce, noncelen);
 
@@ -1625,7 +1707,11 @@ int kcapi_md_destroy(struct kcapi_handle *handle)
  * @key: key buffer - input
  * @keylen: length of key buffer - input
  *
- * This call is applicable for keyed message digests.
+ * With this function, the caller sets the key for subsequent hashing
+ * operations. This call is applicable for keyed message digests.
+ *
+ * After the caller provided the key, the caller may securely destroy the key
+ * as it is now maintained by the kernel.
  *
  * Return: 0 upon success;
  *	   < 0 in case of error
